import aw_client
from datetime import datetime, timedelta, timezone
import socket
import netifaces
import json
import requests
from urllib.parse import urlparse  # Import to help extract domains

# Function to convert UTC to local time
def utc_to_local(utc_dt):
    return utc_dt.astimezone()  # Converts to local time zone

# Function to parse event timestamps
def parse_event_time(event_timestamp):
    if isinstance(event_timestamp, datetime):
        return event_timestamp
    elif isinstance(event_timestamp, str):
        # Remove 'Z' and replace with '+00:00' if present
        if event_timestamp.endswith('Z'):
            event_timestamp = event_timestamp[:-1] + '+00:00'
        return datetime.fromisoformat(event_timestamp)
    else:
        raise TypeError(f"Unsupported type for event['timestamp']: {type(event_timestamp)}")

# Initialize the ActivityWatch client
client = aw_client.ActivityWatchClient("my-client")

# Get the current time and set the start of today (midnight)
now = datetime.now().astimezone()
start_of_today = now.replace(hour=0, minute=0, second=0, microsecond=0)

# Get the buckets for window-related activities, afk, and web
buckets = client.get_buckets()

# Filter relevant buckets
window_buckets = [bucket for bucket in buckets.values() if 'window' in bucket['client']]
afk_buckets = [bucket for bucket in buckets.values() if 'afk' in bucket['id']]
web_buckets = [bucket for bucket in buckets.values() if 'web' in bucket['id']]

# Helper function to extract the date from a timestamp
def extract_day(timestamp):
    timestamp = parse_event_time(timestamp)
    timestamp = utc_to_local(timestamp)
    return timestamp.date().isoformat()

# Helper function to get the hour from a timestamp
def format_datetime(timestamp):
    timestamp = parse_event_time(timestamp)
    timestamp = utc_to_local(timestamp)
    return timestamp.hour

# Get the machine's hostname and non-loopback IP address
def get_machine_info():
    hostname = socket.gethostname()
    ip_address = None

    # Get all network interfaces and their IP addresses
    for interface in netifaces.interfaces():
        addresses = netifaces.ifaddresses(interface)
        if netifaces.AF_INET in addresses:
            for link in addresses[netifaces.AF_INET]:
                ip = link['addr']
                # Skip loopback addresses
                if not ip.startswith("127."):
                    ip_address = ip
                    break
        if ip_address:
            break

    if ip_address is None:
        ip_address = 'Unable to get IP address'

    return hostname, ip_address

# Helper function to calculate AFK sessions and ignore sessions < 5 minutes
def aggregate_afk_sessions(events, start_time):
    sessions_per_day = {}
    for event in events:
        event_time = parse_event_time(event['timestamp'])
        event_time = utc_to_local(event_time)
        if event_time < start_time:
            continue  # Skip events before start of today

        # Convert the duration of the AFK event to minutes
        duration_in_seconds = event['duration'] if isinstance(event['duration'], (int, float)) else event['duration'].total_seconds()
        duration_in_minutes = duration_in_seconds / 60

        # Ignore AFK sessions shorter than 5 minutes
        if duration_in_minutes < 5:
            continue

        # Get the start time hour of the event
        start_time_hour = event_time.hour

        # Parse the day from the timestamp
        day = event_time.date().isoformat()

        # Store the session with start and duration in minutes
        session = {"start": start_time_hour, "minutes": int(duration_in_minutes)}
        if day in sessions_per_day:
            sessions_per_day[day].append(session)
        else:
            sessions_per_day[day] = [session]
    return sessions_per_day

# Helper function to extract the domain from a URL
def get_domain_from_url(url):
    parsed_url = urlparse(url)
    return parsed_url.netloc

# Helper function to get duration in minutes
def get_duration_in_minutes(duration):
    if isinstance(duration, (int, float)):
        duration_in_seconds = duration
    elif isinstance(duration, timedelta):
        duration_in_seconds = duration.total_seconds()
    else:
        raise TypeError(f"Unsupported type for duration: {type(duration)}")
    return duration_in_seconds / 60

# Collect and aggregate app usage time per day
app_usage_per_day = {}
for bucket in window_buckets:
    events = client.get_events(bucket['id'], start=start_of_today)
    for event in events:
        event_time = parse_event_time(event['timestamp'])
        event_time = utc_to_local(event_time)
        if event_time < start_of_today:
            continue  # Skip events before start of today

        app = event['data'].get('app', 'Unknown app')
        duration_in_minutes = get_duration_in_minutes(event['duration'])
        day = event_time.date().isoformat()
        if app not in app_usage_per_day:
            app_usage_per_day[app] = {}
        if day in app_usage_per_day[app]:
            app_usage_per_day[app][day] += duration_in_minutes
        else:
            app_usage_per_day[app][day] = duration_in_minutes

# Collect and aggregate AFK sessions
afk_sessions_per_day = {}
for bucket in afk_buckets:
    events = client.get_events(bucket['id'], start=start_of_today)
    daily_afk_sessions = aggregate_afk_sessions(events, start_of_today)
    for day, sessions in daily_afk_sessions.items():
        if day in afk_sessions_per_day:
            afk_sessions_per_day[day].extend(sessions)
        else:
            afk_sessions_per_day[day] = sessions

# Collect and aggregate web usage time per domain
web_usage_per_day = {}
for bucket in web_buckets:
    events = client.get_events(bucket['id'], start=start_of_today)
    for event in events:
        event_time = parse_event_time(event['timestamp'])
        event_time = utc_to_local(event_time)
        if event_time < start_of_today:
            continue  # Skip events before start of today

        url = event['data'].get('url', 'Unknown URL')
        domain = get_domain_from_url(url)
        duration_in_minutes = get_duration_in_minutes(event['duration'])

        day = event_time.date().isoformat()
        if domain not in web_usage_per_day:
            web_usage_per_day[domain] = {}
        if day in web_usage_per_day[domain]:
            web_usage_per_day[domain][day] += duration_in_minutes
        else:
            web_usage_per_day[domain][day] = duration_in_minutes

# Collect the data for the JSON structure
hostname, ip_address = get_machine_info()
data = {
    'hostname': hostname,
    'ip_address': ip_address
}

# Format app usage data for JSON
n = 0
for app, daily_usage in app_usage_per_day.items():
    for day, duration in daily_usage.items():
        n += 1
        data['appName_' + str(n)] = app
        data['appStart_' + str(n)] = day
        data['appDuration_' + str(n)] = round(duration, 2)
        break  # Only take the first entry per app
data['nApp'] = n

# Format web usage data for JSON
n = 0
for domain, daily_usage in web_usage_per_day.items():
    for day, duration in daily_usage.items():
        n += 1
        data['webDomain_' + str(n)] = domain
        data['webDate_' + str(n)] = day
        data['webDuration_' + str(n)] = round(duration, 2)
        break  # Only take the first entry per domain
data['nWeb'] = n

# Format AFK data for JSON
n = 0
for day, sessions in afk_sessions_per_day.items():
    session = sessions[0]  # Take the first session
    n += 1
    duration = session.get('minutes')
    start = session.get('start')
    data['afkDay_' + str(n)] = day
    data['afkStart_' + str(n)] = start
    data['afkDuration_' + str(n)] = duration
data['nAfk'] = n

# Convert the data to JSON format and print it (temporary feature)
json_output = json.dumps(data, indent=4)
print("Data to be sent:")
print(json_output)

# Send data to the endpoint
url = 'https://panel.qinaya.co/api/mobile.asp?action=monitor2'
response = requests.post(url, data=data)

def main():
    print("Qinaya Rocks!")

if __name__ == '__main__':
    main()

